import { serve } from "https://deno.land/std@0.224.0/http/mod.ts";
import {
  serveDir,
  serveFile,
} from "https://deno.land/std@0.224.0/http/file_server.ts";

import {
  BuzzEvent,
  ClientState,
  RequestPacket,
  ServerState,
  tsToDate,
} from "../frontend/src/API.ts";
import { persistence } from "./src/persistence.ts";
import { apply, getInstanceId } from "./src/processor.ts";
import { handleRequest } from "./src/protocolHandler.ts";
import { Settings, State } from "./src/state.ts";

const HTTP_PORT = 42593;
const FRONTEND_BUILD_DIR = "./frontend-build/";

const state = await persistence.replayTransactions();
const stateStreams: Map<string, WritableStreamDefaultWriter<State>> = new Map();

handleStateUpdate(state);

function reqHandler(req: Request) {
  if (req.headers.get("upgrade") !== "websocket") {
    const pathname = new URL(req.url).pathname;
    if (pathname === "/") {
      return serveFile(req, FRONTEND_BUILD_DIR + "/index.html");
    }
    return serveDir(req, { fsRoot: FRONTEND_BUILD_DIR });
    // return new Response(null, { status: 501 });
  }

  const { socket: ws, response } = Deno.upgradeWebSocket(req);
  console.log("new websocket connection for", req.url);

  const connId = crypto.randomUUID();

  const wsStateStreams: Map<string, WritableStreamDefaultWriter<State>> =
    new Map();

  async function streamProperty<T>(
    lens: (state: State) => T | undefined,
    reqId: number,
    kind: string,
    sendInitial = true,
  ) {
    const streamId = `${connId}-${reqId}-${kind}`;
    const { readable, writable } = new TransformStream<State, State>();
    const writer = writable.getWriter();
    stateStreams.set(streamId, writer);
    wsStateStreams.set(streamId, writer);
    const initial = lens(state);
    let val = JSON.stringify({ id: reqId, streamVal: initial });
    if (sendInitial) {
      try {
        ws.send(val);
      } catch (e) {
        console.error(e);
        // TODO refactor
        stateStreams.delete(streamId);
        wsStateStreams.delete(streamId);
        return;
      }
      if (initial === undefined) {
        stateStreams.delete(streamId);
        wsStateStreams.delete(streamId);
        return;
      }
    }
    for await (const stateSnapshot of readable) {
      if (!stateStreams.has(streamId)) {
        return;
      }
      const streamVal = lens(stateSnapshot);
      const current = JSON.stringify({ id: reqId, streamVal });
      if (val !== current) {
        try {
          ws.send(current);
        } catch (e) {
          console.error(e);
          stateStreams.delete(streamId);
          wsStateStreams.delete(streamId);
          return;
        }
        val = current;
      }
      if (streamVal === undefined) {
        stateStreams.delete(streamId);
        wsStateStreams.delete(streamId);
        return;
      }
    }
  }

  ws.onerror = (_ev) => {
    console.log("websocket errored");
    ws.close();
  };
  ws.onclose = (ev) => {
    console.log("websocket closed", ev.reason);
    for (const reqId of wsStateStreams.keys()) {
      stateStreams.delete(reqId);
    }
  };
  ws.onmessage = (message: MessageEvent<string>) => {
    // NOTE: This doesn't validate the message.
    const req: RequestPacket = JSON.parse(message.data);
    try {
      handlePacket(req);
    } catch (e) {
      console.log("failed to handle packet");
      console.log(req);
      console.error(e);
    }
  };
  return response;

  function handlePacket(req: RequestPacket) {
    switch (req.payload.kind) {
      case "StreamServerState": {
        const instanceId = getInstanceId(state, {
          instanceId: req.payload.instance,
          sessionId: req.payload.session,
        });
        streamProperty<ServerState>(
          (state) => {
            if (!state.instances[instanceId]) {
              return undefined;
            }
            const {
              teams,
              cooldown,
              lastWinner,
              buzzersEnabled,
              enableTimestamp,
              buzzOrder,
            } = state.instances[instanceId];
            return {
              teams: Object.values(teams),
              cooldown: cooldown,
              lastWinner,
              buzzersEnabled,
              instanceId,
              enableTimestamp,
              buzzOrder,
            };
          },
          req.id,
          req.payload.kind,
          true,
        );
        break;
      }
      case "StreamSettings": {
        const instanceId = getInstanceId(state, {
          instanceId: req.payload.instance,
          sessionId: req.payload.session,
        });
        streamProperty<Settings>(
          (state) => state.instances[instanceId]?.settings,
          req.id,
          req.payload.kind,
          true,
        );
        break;
      }
      case "StreamClientState": {
        const sessionId = req.payload.session!;
        const instanceId = getInstanceId(state, { sessionId });
        streamProperty<ClientState>(
          (state) => {
            const instance = state.instances[instanceId];
            if (!instance) {
              return undefined;
            }
            const session = state.sessions[sessionId];
            for (const team of Object.values(instance.teams)) {
              for (const device of Object.values(team.devices)) {
                if (device.uuid === session?.deviceId) {
                  return { ownTeam: team, deviceName: device.name };
                }
              }
            }
          },
          req.id,
          req.payload.kind,
          true,
        );
        break;
      }
      case "StreamBuzzFeed": {
        const instanceId = getInstanceId(state, {
          sessionId: req.payload.session,
          instanceId: req.payload.instance,
        });
        streamProperty<BuzzEvent>(
          (state) => state.instances[instanceId]?.lastBuzz,
          req.id,
          req.payload.kind,
          false,
        );
        break;
      }
      default: {
        const resp = handleRequest(state, req);
        try {
          ws.send(JSON.stringify(resp));
        } catch (e) {
          console.error(e);
        }
        handleStateUpdate(state);
        break;
      }
    }
  }
}

serve((req: Request) => {
  try {
    return reqHandler(req);
  } catch (e) {
    console.error(e);
    return new Response(null, { status: 500 });
  }
}, { port: HTTP_PORT });

function handleStateUpdate(state: State) {
  while (state.scheduledTransactions.length !== 0) {
    const scheduledTransaction = state.scheduledTransactions.shift()!;
    const ts: Date = tsToDate(scheduledTransaction.timestamp);
    const sleepTime = ts.getTime() - new Date().getTime();
    console.log("scheduledTransaction ts:", ts, "sleepTime:", sleepTime);
    setTimeout(() => {
      try {
        apply(state, scheduledTransaction);
      } catch (e) {
        console.log("failed to apply scheduled transaction", ts);
        console.error(e);
      }
      handleStateUpdate(state);
    }, sleepTime > 0 ? sleepTime : 0);
  }

  // Send out updated state to stream listeners.
  const stateSnapshot: State = JSON.parse(JSON.stringify(state));
  // console.log("#stateStreams", stateStreams.size);
  // for (const stateStreamKey of stateStreams.keys()) {
  //   console.log('-', stateStreamKey)
  // }
  for (const streamWriter of stateStreams.values()) {
    streamWriter.write(stateSnapshot);
  }
}
