import { readLines } from "https://deno.land/std@0.224.0/io/mod.ts";
import { apply } from "./processor.ts";

import { emptyState, State } from "./state.ts";
import { Transaction } from "./transactions.ts";

function txHash(tx: Transaction): string {
  const allKeys: Set<string> = new Set();
  JSON.stringify(tx, (key, value) => (allKeys.add(key), value));
  return JSON.stringify(tx, Array.from(allKeys).sort());
}

export class TransactionPersistence {
  private initialized = false;
  private folder: string;
  constructor(folder: string) {
    this.folder = folder;
  }

  persist(state: State, tx: Transaction, instanceId?: string) {
    if (!this.initialized) return;
    if (
      !instanceId ||
      (tx.kind !== "CreateInstance" && !state.instances[instanceId])
    ) return;
    if (tx.kind === "KeepAlive") return;
    const path = this.folder + "/" + instanceSlug(instanceId) + ".jsonl";
    Deno.writeTextFileSync(path, JSON.stringify(tx) + "\n", { append: true });
  }

  async replayTransactions(): Promise<State> {
    const state = emptyState();
    Deno.mkdirSync(this.folder, { recursive: true });
    const appliedTxHashes = new Set();
    for (const dirEntry of Deno.readDirSync(this.folder)) {
      if (dirEntry.isFile && dirEntry.name.endsWith(".jsonl")) {
        console.log(dirEntry.name);
        const file = Deno.openSync(this.folder + "/" + dirEntry.name);
        for await (const line of readLines(file)) {
          const tx: Transaction = JSON.parse(line);
          appliedTxHashes.add(txHash(tx));
          try {
            apply(state, tx);
          } catch (e) {
            console.error(e);
          }
        }
      }
    }
    console.log("live instances:");
    console.log(Object.keys(state.instances));
    state.scheduledTransactions = state.scheduledTransactions.filter((tx) =>
      !appliedTxHashes.has(txHash(tx))
    );
    this.initialized = true;
    return state;
  }

  archiveInstance(instanceId: string) {
    console.log("archiving instance", instanceId);
    const path = this.folder + "/" + instanceSlug(instanceId) + ".jsonl";
    console.log("path:", path);
    // Note that this overwrites existing .bak files:
    Deno.renameSync(path, path + ".bak");
  }
}

export const persistence = new TransactionPersistence("./instances/");

function instanceSlug(s?: string): string | undefined {
  return s?.replaceAll(/[^a-zA-Z0-9_-]/g, ".");
}
