import {
  BuzzTransaction,
  CreateInstanceTransaction,
  Timestamp,
  Transaction,
  TransactionRt,
} from "./transactions.ts";
import { Device, InstanceState, State, Team } from "./state.ts";
import { dateToTs, tsToDate } from "../../frontend/src/API.ts";
import { persistence } from "./persistence.ts";

export class FailedPrecondition extends Error {
  constructor(message = "") {
    super(message);
  }
}

const INSTANCE_DELETE_POLL_INTERVAL_MS = 1_000 * 60 * 60 * 24;

function applyCreateInstance(state: State, tx: CreateInstanceTransaction) {
  if (tx.instanceId in state.instances) {
    throw new FailedPrecondition("Instance already exists!");
  }
  if (tx.sessionId in state.sessions) {
    throw new FailedPrecondition("Session id already exists!");
  }

  const lobbyTeam = {
    uuid: tx.instanceId,
    name: "Lobby",
    excluded: true,
    buzzedDuringCooldown: false,
    devices: {},
    isLobby: true,
    score: 0,
  };

  const settings = {
    timeout: 5,
    rearmAfter: 0,
    autoReenable: false,
    soundEnabled: false,
    allowTeamChanging: true,
    hideDevices: false,
  };

  const instance = {
    uuid: tx.instanceId,
    authKey: tx.authKey,
    teams: { [lobbyTeam.uuid]: lobbyTeam },
    buzzersEnabled: false,
    buzzerEpoch: 0,
    buzzOrder: [],
    settings,
  };

  const session = {
    uuid: tx.sessionId,
    instance: instance.uuid,
    admin: true,
  };

  state.instances[instance.uuid] = instance;
  state.sessions[tx.sessionId] = session;
  const deleteTs = new Date(
    tsToDate(tx.timestamp).getTime() + INSTANCE_DELETE_POLL_INTERVAL_MS,
  );
  state.scheduledTransactions.push({
    kind: "ScheduledDeleteInstance",
    instanceId: instance.uuid,
    timestamp: dateToTs(deleteTs),
  });
}

function applyBuzz(state: State, tx: BuzzTransaction) {
  const { instance, team, session } = getSessionTeam(state, tx);
  if (team.excluded && !session.admin) {
    throw new FailedPrecondition("Team excluded!");
  }
  if (!instance.buzzersEnabled && !session.admin) {
    if (team.uuid !== instance.lastWinner) {
      team.buzzedDuringCooldown = true;
    }
  } else {
    instance.lastWinner = team.uuid;
    instance.buzzersEnabled = false;
    instance.lastBuzz = {
      teamUuid: team.uuid,
      timestamp: tx.timestamp,
    };
    instance.cooldown = instance.settings.timeout;
    state.scheduledTransactions.push({
      kind: "DecrementCountdown",
      instanceId: instance.uuid,
      n: instance.settings.timeout,
      timestamp: tx.timestamp,
      epoch: instance.buzzerEpoch,
    });
  }

  if (!instance.buzzOrder.find((event) => event.teamUuid === team.uuid)) {
    instance.buzzOrder.push({
      teamUuid: team.uuid,
      timestamp: tx.timestamp,
    });
  }
}

function applyRearm(state: State, instanceId: string, timestamp: Timestamp) {
  const instance = state.instances[instanceId];
  if (!instance) throw new FailedPrecondition("Unknown instance.");
  if (!instance.buzzersEnabled) {
    instance.buzzersEnabled = true;
    instance.buzzerEpoch += 1;
    instance.enableTimestamp = timestamp;
    instance.buzzOrder = [];
    for (const team of Object.values(instance.teams)) {
      team.buzzedDuringCooldown = false;
    }
  }
}

function instanceDevices(instance: InstanceState) {
  const devices: { [deviceId: string]: Device } = {};
  for (const team of Object.values(instance.teams)) {
    for (const device of Object.values(team.devices)) {
      devices[device.uuid] = device;
    }
  }
  return devices;
}

/**
 * Apply a transaction to the current state. This either throws a `FailedPrecondition` or mutates the state.
 * @throws {FailedPrecondition}
 */
export function apply(state: State, tx: Transaction) {
  console.log("apply tx:", JSON.stringify(tx));
  // Runtime-check the transaction object to make sure we don't accidentally poison our state.
  const checked = TransactionRt.safeParse(tx);
  if (!checked.success) {
    throw new FailedPrecondition(`Malformed transaction: ${checked.error}`);
  }
  const instanceId = getInstanceId(state, tx);
  persistence.persist(state, tx, instanceId);

  switch (tx.kind) {
    case "CreateInstance":
      applyCreateInstance(state, tx);
      break;
    case "DeleteInstance": {
      const adminSession = state.sessions[tx.sessionId];
      if (!adminSession) throw new FailedPrecondition("Session not found!");
      if (!adminSession.admin) throw new FailedPrecondition("Not admin!");

      delete state.instances[adminSession.instance];
      for (const sessionId of Object.keys(state.sessions)) {
        if (state.sessions[sessionId].instance === adminSession.instance) {
          delete state.sessions[sessionId];
        }
      }
      break;
    }
    case "JoinInstance": {
      if (tx.sessionId in state.sessions) {
        throw new FailedPrecondition("Session id already exists!");
      }
      const instance = state.instances[tx.instanceId];
      if (!instance) throw new FailedPrecondition("Instance not found");
      if (instance.authKey === tx.authKey) {
        // Create admin session
        state.sessions[tx.sessionId] = {
          instance: instance.uuid,
          uuid: tx.sessionId,
          admin: true,
        };
      } else if (tx.authKey) {
        throw new FailedPrecondition("Incorrect AuthKey");
      } else if (tx.name && tx.deviceId) {
        // Add device to team
        if (tx.deviceId in instanceDevices(instance)) {
          throw new FailedPrecondition("Device ID already exists");
        }
        for (const device of Object.values(instanceDevices(instance))) {
          if (tx.name === device.name) {
            throw new FailedPrecondition("Device name already in use");
          }
        }
        instance.teams[tx.instanceId].devices[tx.deviceId] = {
          uuid: tx.deviceId,
          name: tx.name,
          lastSeenTimestamp: tx.timestamp,
        };
        state.sessions[tx.sessionId] = {
          instance: instance.uuid,
          uuid: tx.sessionId,
          deviceId: tx.deviceId,
          admin: false,
        };
      } else {
        throw new FailedPrecondition(
          "Must supply either name and deviceId or authKey",
        );
      }
      break;
    }
    case "CreateTeam": {
      const { instance, session } = getSessionAndInstance(state, tx);
      if (tx.newTeamId in instance.teams) {
        throw new FailedPrecondition("Team ID is already taken!");
      }
      for (const team of Object.values(instance.teams)) {
        if (team.name === tx.name) {
          throw new FailedPrecondition("Team name is already taken!");
        }
      }
      if (!session.admin && !instance.settings.allowTeamChanging) {
        throw new FailedPrecondition("Team creation is currently disabled!");
      }

      const team: Team = {
        uuid: tx.newTeamId,
        name: tx.name,
        excluded: false,
        buzzedDuringCooldown: false,
        devices: {},
        isLobby: false,
        score: 0,
      };
      if (!session.admin) {
        const { team: oldTeam } = getSessionTeam(state, tx);
        instance.teams[tx.newTeamId] = team;
        const deviceId = session.deviceId!;
        team.devices[deviceId] = oldTeam.devices[deviceId];
        delete oldTeam.devices[deviceId];
      } else {
        instance.teams[tx.newTeamId] = team;
      }
      break;
    }
    case "RenameTeam": {
      const { instance, session, team } = getSessionTeam(state, tx);
      if (team.isLobby) {
        throw new FailedPrecondition("Can't change name of lobby!");
      }
      if (!session.admin && !instance.settings.allowTeamChanging) {
        throw new FailedPrecondition("Team renaming is currently disabled!");
      }
      for (const t of Object.values(instance.teams)) {
        if (t.name === tx.name) {
          throw new FailedPrecondition("Team name is already taken!");
        }
      }
      team.name = tx.name;
      break;
    }
    case "SetTeamAltName": {
      const { session, team } = getSessionTeam(state, tx);
      if (team.isLobby) {
        throw new FailedPrecondition("Lobby does not have an alt name");
      }
      if (!session.admin) {
        throw new FailedPrecondition("Not allowed!");
      }
      team.altName = tx.name;
      break;
    }
    case "RenameDevice": {
      const { instance, device } = getSessionDevice(state, tx);
      for (const team of Object.values(instance.teams)) {
        for (const device of Object.values(team.devices)) {
          if (device.name === tx.name) {
            throw new FailedPrecondition("Device name is already taken!");
          }
        }
      }
      device.name = tx.name;
      break;
    }
    case "JoinTeam": {
      const { instance, team: oldTeam } = getSessionTeam(state, tx);
      const { device } = getSessionDevice(state, tx);
      if (!instance.teams[tx.newTeamId]) {
        throw new FailedPrecondition("Unknown target team");
      }
      if (!instance.settings.allowTeamChanging) {
        throw new FailedPrecondition("Team changing is not currently allowed!");
      }
      if (oldTeam.uuid === tx.newTeamId) {
        throw new FailedPrecondition(
          "Try joining a team you're not currently on.",
        );
      }

      instance.teams[tx.newTeamId].devices[device.uuid] =
        oldTeam.devices[device.uuid];
      delete oldTeam.devices[device.uuid];
      break;
    }
    case "UpdateSettings": {
      const { instance, session } = getSessionAndInstance(state, tx);
      if (!session.admin) throw new FailedPrecondition("Permission denied.");

      const {
        timeout,
        rearmAfter,
        autoReenable,
        soundEnabled,
        allowTeamChanging,
        hideDevices,
      } = tx;
      instance.settings = {
        timeout,
        rearmAfter,
        autoReenable,
        soundEnabled,
        allowTeamChanging,
        hideDevices,
      };
      break;
    }
    case "SetBuzzersEnabled": {
      const { session, instance } = getSessionAndInstance(state, tx);
      if (!session.admin) throw new FailedPrecondition("Permission denied.");

      if (tx.enabled && instance.cooldown !== undefined) {
        instance.cooldown = undefined;
      }
      if (tx.enabled) {
        applyRearm(state, instance.uuid, tx.timestamp);
      } else {
        instance.lastWinner = undefined;
        instance.buzzersEnabled = false;
      }
      break;
    }
    case "ScheduledRearm": {
      const instance = state.instances[tx.instanceId];
      if (instance && instance.cooldown !== undefined) break;
      if (instance.buzzerEpoch !== tx.epoch) return;
      applyRearm(state, tx.instanceId, tx.timestamp);
      break;
    }
    case "DeleteTeam": {
      const { instance, team } = getSessionTeam(state, tx);
      if (team.isLobby) throw new FailedPrecondition("Can't delete lobby!");
      // Note: This allows users to move others to the lobby by joining and deleting their team.
      // Should we gate this to only allow users to delete empty teams?

      const lobby = instance.teams[instance.uuid];
      for (const device of Object.values(team.devices)) {
        lobby.devices[device.uuid] = device;
      }
      delete instance.teams[team.uuid];
      break;
    }
    case "SetPoints": {
      const { session, team } = getSessionTeam(state, tx);
      if (!session.admin) throw new FailedPrecondition("Permission denied.");
      team.score = tx.points;
      break;
    }
    case "ResetPoints": {
      const { session, instance } = getSessionAndInstance(state, tx);
      if (!session.admin) throw new FailedPrecondition("Permission denied.");
      for (const team of Object.values(instance.teams)) {
        team.score = 0;
      }
      break;
    }
    case "ChangeExclusion": {
      const { session, team } = getSessionTeam(state, tx);
      if (!session.admin) throw new FailedPrecondition("Permission denied.");
      team.excluded = tx.excluded;
      break;
    }
    case "Buzz": {
      applyBuzz(state, tx);
      break;
    }
    case "BuzzRandomTeam": {
      const { instance } = getSessionAndInstance(state, tx);
      const teams = Object.values(instance.teams).filter((team) =>
        !team.isLobby && !team.excluded
      );
      if (teams.length === 0) {
        throw new FailedPrecondition("Can't find a team to buzz.");
      }
      const teamId = teams[~~(Math.random() * teams.length)].uuid;
      applyBuzz(state, {
        kind: "Buzz",
        sessionId: tx.sessionId,
        timestamp: tx.timestamp,
        teamId,
      });
      break;
    }
    case "DecrementCountdown": {
      const instance = state.instances[tx.instanceId];
      if (!instance) throw new FailedPrecondition("Instance not found!");
      if (instance.buzzerEpoch !== tx.epoch) return;

      if (instance.cooldown === undefined && instance.buzzersEnabled) {
        // No-op
      } else if (tx.n <= 0) {
        instance.cooldown = undefined;
        // instance.lastWinner = undefined;
        if (instance.settings.autoReenable) {
          const delta = instance.settings.rearmAfter * 1_000;
          const timestamp = dateToTs(new Date(new Date().getTime() + delta));
          state.scheduledTransactions.push({
            kind: "ScheduledRearm",
            instanceId: tx.instanceId,
            timestamp,
            epoch: instance.buzzerEpoch,
          });
        }
      } else {
        instance.cooldown = tx.n;
        state.scheduledTransactions.push({
          kind: "DecrementCountdown",
          instanceId: tx.instanceId,
          n: tx.n - 1,
          timestamp: dateToTs(new Date(new Date().getTime() + 1_000)),
          epoch: instance.buzzerEpoch,
        });
      }
      break;
    }
    case "KeepAlive": {
      const { session, team } = getSessionTeam(state, tx);
      if (!session.deviceId) {
        throw new FailedPrecondition("No associated device.");
      }
      team.devices[session.deviceId].lastSeenTimestamp = tx.timestamp;
      break;
    }
    case "ScheduledDeleteInstance": {
      const instance = state.instances[tx.instanceId];
      if (!instance) throw new FailedPrecondition("Unknown instance.");
      const txTs = tsToDate(tx.timestamp);
      if (
        instance.lastBuzz &&
        txTs.getTime() - tsToDate(instance.lastBuzz.timestamp).getTime() <
        INSTANCE_DELETE_POLL_INTERVAL_MS
      ) {
        // There's been some (relevant) activity in the last 24h. Let's postpone instance deletion.
        state.scheduledTransactions.push({
          kind: "ScheduledDeleteInstance",
          instanceId: tx.instanceId,
          timestamp: dateToTs(
            new Date(txTs.getTime() + INSTANCE_DELETE_POLL_INTERVAL_MS),
          ),
        });
      } else {
        // We can safely delete the instance.
        for (const session of Object.values(state.sessions)) {
          if (session.instance === instance.uuid) {
            delete state.sessions[session.uuid];
          }
        }
        delete state.instances[tx.instanceId];
        persistence.archiveInstance(tx.instanceId);
      }
      break;
    }
  }
}

function getSessionAndInstance(
  state: State,
  { sessionId }: { sessionId: string },
) {
  const session = state.sessions[sessionId];
  if (!session) {
    throw new FailedPrecondition("Session not found");
  }
  const instance = state.instances[session.instance];
  if (!instance) {
    throw Error("Internal error: Instance not found");
  }
  return { instance, session };
}

function getSessionTeam(
  state: State,
  tx: { sessionId: string; teamId?: string; deviceId?: string },
) {
  const { instance, session } = getSessionAndInstance(state, tx);
  const deviceId = tx.deviceId || session.deviceId;
  if (tx.teamId && deviceId === undefined) {
    if (!session.admin) {
      throw new FailedPrecondition(
        "Modifying arbitrary teams is only allowed for admins.",
      );
    }
    return { instance, session, team: instance.teams[tx.teamId] };
  } else {
    for (const team of Object.values(instance.teams)) {
      for (const device of Object.values(team.devices)) {
        if (device.uuid === deviceId) {
          return { instance, session, team };
        }
      }
    }
    throw Error("Internal error: Session without associated team");
  }
}

function getSessionDevice(
  state: State,
  tx: { sessionId: string; deviceId?: string },
) {
  const { instance, session } = getSessionAndInstance(state, tx);
  if (tx.deviceId) {
    if (!session.admin) {
      throw new FailedPrecondition(
        "Modifying arbitrary devices is only allowed for admins.",
      );
    }
    for (const team of Object.values(instance.teams)) {
      if (team.devices[tx.deviceId]) {
        return { instance, session, device: team.devices[tx.deviceId] };
      }
    }
    throw new FailedPrecondition("Device id not found");
  } else {
    for (const team of Object.values(instance.teams)) {
      for (const device of Object.values(team.devices)) {
        if (device.uuid === session.deviceId) {
          return { instance, session, device };
        }
      }
    }
    throw Error("Internal error: Session without associated team");
  }
}

export function getInstanceId(
  state: State,
  tx: { sessionId?: string; instanceId?: string },
) {
  if (tx.instanceId) {
    return tx.instanceId;
  }
  if (!tx.sessionId) {
    throw new Error("tx with neither sessionId or instanceId");
  }
  return state.sessions[tx.sessionId]?.instance || "";
}
