import {
  dateToTs,
  RequestPacket,
  ResponsePacket,
} from "../../frontend/src/API.ts";
import { apply, FailedPrecondition } from "./processor.ts";
import { State, uuid } from "./state.ts";

export function handleRequest(
  state: State,
  req: RequestPacket,
): ResponsePacket {
  const tsNow = dateToTs(new Date());
  try {
    const kind = req.payload.kind;
    switch (kind) {
      case "StreamServerState":
      case "StreamSettings":
      case "StreamClientState":
      case "StreamBuzzFeed":
        throw new FailedPrecondition("unreachable");

      case "SpawnInstance": {
        const sessionId = uuid();
        apply(state, {
          kind: "CreateInstance",
          instanceId: req.payload.instance,
          authKey: req.payload.authKey,
          sessionId,
          timestamp: tsNow,
        });
        return {
          id: req.id,
          response: {
            kind: "SpawnInstance",
            session: sessionId,
          },
        };
      }
      case "JoinInstance": {
        const sessionId = uuid();
        const deviceId = uuid();
        apply(state, {
          kind,
          instanceId: req.payload.instance,
          name: req.payload.name,
          authKey: req.payload.authKey,
          timestamp: tsNow,
          sessionId,
          deviceId,
        });
        return {
          id: req.id,
          response: {
            kind: "JoinInstance",
            session: sessionId,
            // device: deviceId
          },
        };
      }

      case "KeepAlive":
        if (req.payload.session !== undefined) {
          apply(state, {
            kind,
            sessionId: req.payload.session,
            timestamp: tsNow,
          });
        }
        break;
      case "SetBuzzersEnabled":
        apply(state, {
          kind,
          sessionId: req.payload.session,
          enabled: req.payload.enabled,
          timestamp: tsNow,
        });
        break;
      case "UpdateSettings":
        apply(state, {
          kind,
          sessionId: req.payload.session,
          timeout: req.payload.settings.timeout,
          rearmAfter: req.payload.settings.rearmAfter,
          autoReenable: req.payload.settings.autoReenable,
          soundEnabled: req.payload.settings.soundEnabled,
          allowTeamChanging: req.payload.settings.allowTeamChanging,
          hideDevices: req.payload.settings.hideDevices,
        });
        break;
      case "DeleteInstance":
        apply(state, { kind, sessionId: req.payload.session });
        break;
      case "BuzzRandomTeam":
        apply(state, {
          kind,
          sessionId: req.payload.session,
          timestamp: tsNow,
        });
        break;
      case "ResetPoints":
        apply(state, { kind, sessionId: req.payload.session });
        break;
      case "Buzz":
        apply(state, {
          kind,
          sessionId: req.payload.session,
          teamId: req.payload.teamId,
          timestamp: tsNow,
        });
        break;
      case "SetPoints":
        apply(state, {
          kind,
          sessionId: req.payload.session,
          teamId: req.payload.teamId,
          points: req.payload.points,
        });
        break;
      case "DeleteTeam":
        apply(state, {
          kind,
          sessionId: req.payload.session,
          teamId: req.payload.teamId,
        });
        break;
      case "RenameDevice":
        apply(state, {
          kind,
          sessionId: req.payload.session,
          name: req.payload.newName,
          deviceId: req.payload.deviceUuid,
        });
        break;
      case "RenameTeam":
        apply(state, {
          kind,
          sessionId: req.payload.session,
          name: req.payload.newName,
          teamId: req.payload.teamId,
        });
        break;
      case "CreateTeam":
        apply(state, {
          kind,
          sessionId: req.payload.session,
          name: req.payload.name,
          newTeamId: uuid(),
        });
        break;
      case "ChangeExclusion":
        apply(state, {
          kind,
          sessionId: req.payload.session,
          teamId: req.payload.teamId,
          excluded: req.payload.excluded,
        });
        break;
      case "JoinTeam":
        apply(state, {
          kind,
          sessionId: req.payload.session,
          newTeamId: req.payload.teamId,
          deviceId: req.payload.deviceUuid,
        });
        break;
    }
    return { id: req.id, response: { kind: "success" } };
  } catch (e) {
    if (e instanceof FailedPrecondition) {
      console.log("couldn't apply transaction: " + e.message);
      return { id: req.id, error: e.message };
    } else {
      console.log("internal while applying transaction: " + e.message);
      // TODO: downgrade to non-nonthrowing handler?
      // malformed requests stop the eventloop with this:
      throw e;
    }
  }
}
